//
// outvo.c
//
// NASM output driver for CA-Visual Objects repository object code
//
// Sabo                                                      09/05/99
//*********************************************************************

#ifdef OF_VO

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <cavoadam.h>

#include "nasm.h"
#include "nasmlib.h"
#include "outform.h"
#include "nasm4vo.h"

#pragma warning(disable : 4996)

static efunc error;
static char vo_infile[FILENAME_MAX];
static char *vo_stdmac[] = {
    "%define __SECT__ [section .text]",
    "%macro __NASM_CDecl__ 1",
    "%endmacro",
    NULL
};
static char *lpchErrors;
static unsigned int cbErrSize, cbErrLen;
static fWell;

typedef struct _SECTION
{	
	char			achName[9];
	unsigned int	uiIndex;
	unsigned char 	*lpbData;
	unsigned long	cbSize;			// Total size
	unsigned long	cbLen;			// Used size
	unsigned char 	*lpbFixups;
	unsigned long	cbFixSize;		// Total size
	unsigned long	cbFixLen;		// Used size
} SECTION;

// VO code can only end up in two different sections: .text and .data
#define SECTION_TEXT	0
#define SECTION_DATA	1
						 
SECTION aSec[2];

#define	SEC_BUF_STEPSIZE	0x1000
#define	ERR_BUF_STEPSIZE	0x512

static void vo_std_init(FILE *fp, efunc errfunc, ldfunc ldef, evalfunc eval) 
{
	memset(&aSec, '\0', sizeof(aSec));	

    error = errfunc;

	cbErrSize = cbErrLen = 0;
	lpchErrors = NULL;

	fWell = TRUE;
}

static int vo_set_info(enum geninfo type, char **val)
{
    return 0;
}


static long vo_add_reloc(int segTo, unsigned int uiLocOff, int iTargetSec, unsigned int uiTargetOff, BOOL fRel)
{
	UINT		ui;
	LPBYTE		lpb;
	DATAFIXUP	df;
	memset(&df, 0, sizeof(DATAFIXUP));
	if(aSec[segTo].cbFixLen + sizeof(DATAFIXUP) > aSec[segTo].cbFixSize)
	{
		if(!(lpb = realloc(aSec[segTo].lpbFixups, aSec[segTo].cbFixSize + SEC_BUF_STEPSIZE)))
		{
			return(0);
		}

		aSec[segTo].lpbFixups = lpb;
		aSec[segTo].cbFixSize += SEC_BUF_STEPSIZE;
	}

	if(iTargetSec != SEG_REPO)
	{
		df.dwLocation	= uiLocOff;
		df.dwTarget		= (unsigned long)ridCur;
		df.dwDisp		= uiTargetOff;
		df.bLocType		= fRel ? LT_RELPOINTER : LT_ABSPOINTER;
		df.bTargetType	= iTargetSec ? TT_DATA : TT_CODE;
	}
	else
	{
		if(EntQueryBlockType((RID)uiTargetOff, &ui) == SUCCESS)
		{
			df.dwLocation	= uiLocOff;
			df.dwTarget		= uiTargetOff;
			df.dwDisp		= 0;
			df.bLocType		= fRel ? LT_RELPOINTER : LT_ABSPOINTER;
			df.bTargetType	= ui == BL_GLOBAL ? TT_DATA : TT_CODE;

			EntBuildDep((RID)uiTargetOff, ridCur);
		}
	}

	memcpy(aSec[segTo].lpbFixups + aSec[segTo].cbFixLen, &df, sizeof(DATAFIXUP));
	aSec[segTo].cbFixLen += sizeof(DATAFIXUP);

	return(0);
}

static void vo_out (long segto, void *data, unsigned long type, long segment, long wrt) 
{
	unsigned long cb;

    if (wrt != NO_SEG) 
	{
		wrt = NO_SEG;		       /* continue to do _something_ */
		error (ERR_NONFATAL, "WRT not supported by COFF output formats");
    }

	cb	= type & OUT_SIZMASK;
	type= type & OUT_TYPMASK;

	if (segto == NO_SEG) 
	{
		if (type != OUT_RESERVE)
		{
			error (ERR_NONFATAL, "attempt to assemble code in [ABSOLUTE] space");
		}
		return;
    }

	if(segto < 0 || segto > 2)
	{
		error (ERR_PANIC, "attempt to write to invalid section");
	}

	segto--;

	while(aSec[segto].cbLen + cb > aSec[segto].cbSize)
	{
		unsigned char	*lpb;

		if(!(lpb = realloc(aSec[segto].lpbData, aSec[segto].cbSize + SEC_BUF_STEPSIZE)))
		{
			return;
		}

		aSec[segto].lpbData = lpb;
		memset(aSec[segto].lpbData + aSec[segto].cbSize, '\0', SEC_BUF_STEPSIZE);
		aSec[segto].cbSize += SEC_BUF_STEPSIZE;
	}

	switch(type)
	{
		case OUT_RESERVE:

			aSec[segto].cbLen += cb;
			break;

		case OUT_RAWDATA:

			memcpy(aSec[segto].lpbData + aSec[segto].cbLen, data, cb);
			aSec[segto].cbLen += cb;
			break;

		case OUT_ADDRESS:

			if (cb != 4 && (segment != NO_SEG || wrt != NO_SEG))
			{
				error(ERR_NONFATAL, "Non-32-bit relocations not supported");
			}
			else 
			{
				if (segment != NO_SEG || wrt != NO_SEG) 
				{
					if (wrt != NO_SEG) 
					{
						error(ERR_NONFATAL, "WRT types not supported");
					}
					else if (segment % 2)
					{
						error(ERR_NONFATAL, "Segment base references not supported");
					}
					else
					{
						vo_add_reloc(segto, aSec[segto].cbLen, segment, *(unsigned long *)data, FALSE);
						*(unsigned long *)data = 0;
					}
				}

				memcpy(aSec[segto].lpbData + aSec[segto].cbLen, data, cb);
				aSec[segto].cbLen += cb;
			}
			break;

		case OUT_REL2ADR:
			error(ERR_NONFATAL, "16-bit relocations not supported");
			break;

		case OUT_REL4ADR:

			if (segment == segto)
			{
				error(ERR_PANIC, "intra-segment OUT_REL4ADR");
			}
			else if (segment == NO_SEG)
			{
				error(ERR_NONFATAL, "Win32 COFF does not correctly support relative references to absolute addresses");
			}
			else 
			{
				if (segment != NO_SEG && segment % 2) 
				{
					error(ERR_NONFATAL, "segment base references not supported");
				}
				else
				{
					vo_add_reloc(segto, aSec[segto].cbLen, segment, *(unsigned long *)data, TRUE);
					*(unsigned long *)data = 0;
				}
				memcpy(aSec[segto].lpbData + aSec[segto].cbLen, data, cb);
				aSec[segto].cbLen += cb;
			}
			break;
	}

	return;
}


static void vo_deflabel (char *name, long segment, long offset, int is_global, char *special) 
{

}


static long vo_section_names (char *name, int pass, int *bits) 
{
	long	i;

	if(name && *name)
	{
		i = !strcmp(name, ".text") ? SECTION_TEXT : SECTION_DATA;

		if(!aSec[i].uiIndex)
		{
			strcpy(aSec[i].achName, name);
			aSec[i].uiIndex = i + 1;

			aSec[i].lpbData = malloc(SEC_BUF_STEPSIZE);
			memset(aSec[i].lpbData, '\0', SEC_BUF_STEPSIZE);
			aSec[i].cbSize = SEC_BUF_STEPSIZE;

			aSec[i].lpbFixups = malloc(SEC_BUF_STEPSIZE);
			aSec[i].cbFixSize = SEC_BUF_STEPSIZE;
		}

		i = aSec[i].uiIndex;
	}
	else
	{
		i = 0;
	}

	return(i);
}

static long vo_segbase (long segment) 
{
    return segment;
}

static int vo_directives (char *directive, char *value, int pass) 
{
    return 0;
}

static void vo_std_filename (char *inname, char *outname, efunc error) 
{
    strcpy(vo_infile, inname);
    standard_extension (inname, outname, ".vo", error);
}

static void vo_cleanup(int ridEnt) 
{
	RID		rid;
	RETCODE rc;
	
	rid = (RID)ridEnt;

	if(aSec[0].uiIndex)
	{
		if(fWell)
		{
			rc = EntSetData(rid, DT_CODE, aSec[0].lpbData, aSec[0].cbLen, aSec[0].cbFixLen ? aSec[0].lpbFixups : NULL, aSec[0].cbFixLen);
				ERRCHK(Error1);

/* Method support requires much more work, so forget about it
			{
				RID		ridClass, ridMod
				UINT	ui;
				NANATOM	atomName;
				LPSTR	lpstr;
				LPCHAR	p;
				CHAR	achBuf[128];

				// If it is a method build dependancy between class and method
				rc = EntQueryBlockType(ridEnt, &ui);
					ERRCHK(Error1);

				if(	ui == BL_METHOD ||
					ui == BL_ACCESS ||
					ui == BL_ASSIGN )
				{
					rc = EntQueryName(ridEnt, &atomName);
						ERRCHK(Error1);

					if((lpstr = SysGetAtomName(atomName)) && (p = strchr(lpstr, ':')) && p - lpstr < sizeof(achBuf))
					{
						strncpy(achBuf, lpstr, p - lpstr);

						rc = EntQueryModule(ridEnt, &ridMod);
							ERRCHK(Error1);

						rc = EntFind(ridMod, SysAddAtom(achBuf), ETYPE_CAIN, &ridClass);
							ERRCHK(Error1);

						if(ridClass)
						{
							EntBuildDep(ridClass, ridEnt);
						}
					}
				}
			}
*/
		}

		rc = EntSetEntry(rid, fWell ? ET_RUN : ENTRY_INVALID, 0);
			ERRCHK(Error2);

		rc = EntSetEntry(rid, fWell ? ET_END : ENTRY_INVALID, aSec[0].cbLen);
			ERRCHK(Error2);

		free(aSec[0].lpbData);
		free(aSec[0].lpbFixups);
	}

	if(aSec[1].uiIndex)
	{
		if(fWell)
		{
			rc = EntSetData(rid, DT_DATA, aSec[1].lpbData, aSec[1].cbLen,  aSec[1].cbFixLen ? aSec[1].lpbFixups : NULL, aSec[1].cbFixLen);
				ERRCHK(Error2);
		}

		free(aSec[1].lpbData);
		free(aSec[1].lpbFixups);
	}

	memset(&aSec, '\0', sizeof(aSec));	

	cbErrSize = 0;

	if(fWell)
	{
		EntSetVit(rid, VIT_WELL);
	}

	return;

Error2:
	EntSetData(rid, DT_CODE, NULL, 0, NULL, 0);
Error1:
	if(aSec[0].uiIndex)
	{
		free(aSec[0].lpbData);
		free(aSec[0].lpbFixups);
	}

	if(aSec[1].uiIndex)
	{
		free(aSec[1].lpbData);
		free(aSec[1].lpbFixups);
	}

	EntSetEntry(rid, ET_RUN, ENTRY_INVALID);
	EntSetEntry(rid, ET_END, ENTRY_INVALID);

	memset(&aSec, '\0', sizeof(aSec));	

	cbErrSize = cbErrLen = 0;
	lpchErrors = NULL;

	return;
}

struct ofmt ofmt_vo = {
    "CA-Visual Objects 2.5a",
    OF_NAME_VO,
    NULL,
    null_debug_arr,
    &null_debug_form,
    vo_stdmac,
    vo_std_init,
    vo_set_info,
    vo_out,
    vo_deflabel,
    vo_section_names,
    vo_segbase,
    vo_directives,
    vo_std_filename,
    vo_cleanup
};


#endif // OF_VO

void vo_report_error(int severity, char *fmt, va_list ap, int iLine, char *lpchFile)
{
	unsigned int	cb;
	char			*p;
	char			achFormat[512];
	char			achErrBuffer[512];

	if ((severity & ERR_MASK) == ERR_WARNING)
	{
		p = "warning";
	}
	else
	{
		fWell = FALSE;
		p = "error";
	}

	sprintf(achFormat, "%s (%u) : %s xxx : %s\r\n", lpchFile, iLine - 4 + iLineCorr, p, fmt);
    vsprintf(achErrBuffer, achFormat, ap);

	cb = strlen(achErrBuffer) + 1;

	while(cbErrLen + cb > cbErrSize)
	{
		if(!cbErrSize)
		{
			// Must use HeapAlloc here, Adam will call HeapFree
			lpchErrors = HeapAlloc(GetProcessHeap(), 0, ERR_BUF_STEPSIZE);
		}
		else
		{
			lpchErrors = HeapReAlloc(GetProcessHeap(), 0, lpchErrors, cbErrSize + ERR_BUF_STEPSIZE);
		}

		cbErrSize += ERR_BUF_STEPSIZE;
	}

	strcpy(lpchErrors + cbErrLen, achErrBuffer);
	cbErrLen += cb - 1;
}

void vo_get_error_info(char **lplpchErrors, unsigned int *lpuiErrors)
{
	*lplpchErrors	= lpchErrors;

	// Last error line must not be terminated by \r\n, but by \0
	if(lpchErrors)
	{
		lpchErrors[cbErrLen-2] = '\0';
		*lpuiErrors		= cbErrLen - 1;
	}
	else
	{
		*lpuiErrors = 0;
	}
	
}

/*
** That's all Folks!
 */